/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Philippe Merle
 */

package org.ow2.frascati.factory.core.instance.implementation.spring;

import java.io.IOException;
import java.util.Locale;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.util.Fractal;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.beans.factory.config.AutowireCapableBeanFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.MessageSourceResolvable;
import org.springframework.context.NoSuchMessageException;
import org.springframework.core.io.Resource;

/**
 * Is a Spring {@link ApplicationContext} where beans are obtained by
 * requesting internal interfaces of a Fractal component.
 *
 * @author <a href="mailto:Philippe.Merle@inria.fr">Philippe Merle</a>
 */

public class ParentApplicationContext implements ApplicationContext
{
  // ======================================================================
  //
  // Internal state.
  //
  // ======================================================================

  /**
   * Logger for this class.
   */
  protected java.util.logging.Logger log = java.util.logging.Logger
    .getLogger(this.getClass().getCanonicalName());

  /**
   * The Fractal component where to search Spring beans.
   */
  protected Component component;

  // ======================================================================
  //
  // Constructor.
  //
  // ======================================================================

  /**
   * Creates a {@link ParentApplicationContext} instance using a Fractal {@link Component}.
   *
   * @param component The Fractal {@link Component} to use.
   */
  public ParentApplicationContext (final Component component)
  {
    this.component = component;
  }

  // ======================================================================
  //
  // Implementation of org.springframework.context.ApplicationContext
  //
  // ======================================================================

  /**
   * @see ApplicationContext#getId()
   */
  public String getId ()
  {
    log.finer("Spring parent context - getId called");
    return null;
  }

  /**
   * @see ApplicationContext#getBean(String)
   */
  public Object getBean (final String name)
    throws BeansException
  {
    return getBean(name, (Class)null);
  }

  /**
   * @see ApplicationContext#getBean(String,Object[])
   */
  public Object getBean (final String name, final Object[] args)
    throws BeansException
  {
    return getBean(name, (Class)null);
  }

  /**
   * @see ApplicationContext#getBean(String,Class)
   */
  public Object getBean (final String name, final Class requiredType)
    throws BeansException
  {
    log.finer("Spring parent context - getBean called for name: " + name);

    // Try to find the requested Spring bean as an internal interface of the Fractal component.
    try {
      Object bean = Fractal.getContentController(component).getFcInternalInterface(name);
      // TODO: Check if bean is instance of requiredType!
      return bean;
    } catch(NoSuchInterfaceException nsie) {
    }

    // TODO: The requested bean is perhaps a property from the SCA composite file.

    // When not found then throw an exception.
    throw new NoSuchBeanDefinitionException("Unable to find Bean with name " + name);
  }

  /**
   * @see ApplicationContext#containsBean(String)
   */
  public boolean containsBean (final String name)
  {
    log.finer("Spring parent context - containsBean called for name: " + name);
    return false;
  }

  /**
   * @see ApplicationContext#isSingleton(String)
   */
  public boolean isSingleton (final String name)
    throws NoSuchBeanDefinitionException
  {
    log.finer("Spring parent context - isSingleton called for name: " + name);
    return false;
  }

  /**
   * @see ApplicationContext#isTpeMatch(String,Class)
   */
  public boolean isTypeMatch (final String name, final Class targetType)
    throws NoSuchBeanDefinitionException
  {
    log.finer("Spring parent context - isTypeMatch called for name: " + name);
    throw new UnsupportedOperationException();
  }

  /**
   * @see ApplicationContext#getType(String)
   */
  public Class getType (final String name)
    throws NoSuchBeanDefinitionException
  {
    log.finer("Spring parent context - getType called for name: " + name);
    return null;
  }

  /**
   * @see ApplicationContext#getAliases(String)
   */
  public String[] getAliases (final String name)
    throws NoSuchBeanDefinitionException
  {
    log.finer("Spring parent context - getAliases called for name: " + name);
    return new String[0];
  }

  /**
   * @see ApplicationContext#getParent()
   */
  public ApplicationContext getParent ()
  {
    log.finer("Spring parent context - getParent called");
    return null;
  }

  /**
   * @see ApplicationContext#getAutowireCapableBeanFactory()
   */
  public AutowireCapableBeanFactory getAutowireCapableBeanFactory ()
    throws IllegalStateException
  {
    log.finer("Spring parent context - getAutowireCapableBeanFactory called");
    return null;
  }

  /**
   * @see ApplicationContext#getDisplayName()
   */
  public String getDisplayName ()
  {
    log.finer("Spring parent context - getDisplayName called");
    return "NoDisplayName";
  }

  /**
   * @see ApplicationContext#getStartupDate()
   */
  public long getStartupDate ()
  {
    log.finer("Spring parent context - getStartupDate called");
    return 0;
  }

  /**
   * @see ApplicationContext#containsBeanDefinition(String)
   */
  public boolean containsBeanDefinition (final String beanName)
  {
    log.finer("Spring parent context - containsBeanDefinition called for name: " + beanName);
    return false;
  }

  /**
   * @see ApplicationContext#getBeanDefinitionCount()
   */
  public int getBeanDefinitionCount ()
  {
    log.finer("Spring parent context - getBeanDefinitionCount called");
    return 0;
  }

  /**
   * @see ApplicationContext#getBeanDefinitionNames()
   */
  public String[] getBeanDefinitionNames ()
  {
    log.finer("Spring parent context - getBeanDefinitionNames called");
    return new String[0];
  }

  /**
   * @see ApplicationContext#getBeanNamesForType(Class)
   */
  public String[] getBeanNamesForType (final Class type)
  {
    log.finer("Spring parent context - getBeanNamesForType called");
    return new String[0];
  }

  /**
   * @see ApplicationContext#getBeanNamesForType(Class,boolean,boolean)
   */
  public String[] getBeanNamesForType (final Class type, final boolean includePrototypes, final boolean includeFactoryBeans)
  {
    log.finer("Spring parent context - getBeanNamesForType called");
    return new String[0];
  }

  /**
   * @see ApplicationContext#getBeansOfType(Class)
   */
  public Map getBeansOfType (final Class type)
    throws BeansException
  {
    log.finer("Spring parent context - getBeansOfType called");
    return null;
  }

  /**
   * @see ApplicationContext#getBeansOfType(Class,boolean,boolean)
   */
  public Map getBeansOfType (final Class type, final boolean includePrototypes, final boolean includeFactoryBeans)
    throws BeansException
  {
    log.finer("Spring parent context - getBeansOfType called");
    return null;
  }

  /**
   * @see ApplicationContext#isPrototype(String)
   */
  public boolean isPrototype (final String theString)
  {
    log.finer("Spring parent context - isPrototype called");
    return false;
  }

  /**
   * @see ApplicationContext#getParentBeanFactory()
   */
  public BeanFactory getParentBeanFactory ()
  {
    log.finer("Spring parent context - getParentBeanFactory called");
    return null;
  }

  /**
   * @see ApplicationContext#containsLocalBean(String)
   */
  public boolean containsLocalBean (final String name)
  {
    log.finer("Spring parent context - containsLocalBean called for name: " + name);
    return false;
  }

  /**
   * @see ApplicationContext#getMessage(String,Object[],String,Locale)
   */
  public String getMessage (final String code, final Object[] args, final String defaultMessage, final Locale locale)
  {
    log.finer("Spring parent context - getMessage called");
    return null;
  }

  /**
   * @see ApplicationContext#getMessage(String,Object[],Locale)
   */
  public String getMessage (final String code, final Object[] args, final Locale locale)
    throws NoSuchMessageException
  {
    log.finer("Spring parent context - getMessage called");
    return null;
  }

  /**
   * @see ApplicationContext#getMessage(MessageSourceResolvable,Locale)
   */
  public String getMessage (final MessageSourceResolvable resolvable, final Locale locale)
    throws NoSuchMessageException
  {
    log.finer("Spring parent context - getMessage called");
    return null;
  }

  /**
   * @see ApplicationContext#publishEvent(ApplicationEvent)
   */
  public void publishEvent (final ApplicationEvent event)
  {
    log.finer("Spring parent context - publishEvent called");
  }

  /**
   * @see ApplicationContext#getResources(String)
   */
  public Resource[] getResources (final String locationPattern)
    throws IOException
  {
    log.finer("Spring parent context - getResources called");
    return new Resource[0];
  }

  /**
   * @see ApplicationContext#getResource(String)
   */
  public Resource getResource (final String location)
  {
    log.finer("Spring parent context - getResource called");
    return null;
  }

  /**
   * @see ApplicationContext#getClassLoader()
   */
  public ClassLoader getClassLoader ()
  {
    log.finer("Spring parent context - getClassLoader called");
    return this.getClass().getClassLoader();
  }
}
