/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Damien Fournier
 */

package org.ow2.frascati.factory.core.parser.util;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URI;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.common.util.Diagnostic;
import org.eclipse.emf.ecore.util.Diagnostician;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.stp.sca.Composite;
import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.ow2.frascati.factory.core.instance.runtime.RuntimeFactory;
import org.ow2.frascati.factory.core.parser.Sca2Ecore;

/**
 * Process composite and java file to add missing elements to the model
 * 
 * @author <a href="mailto:damien.fournier@inria.fr">Damien Fournier</a>
 * @version $Revision$
 */

@Component
public class ModelResolver implements
    org.eclipse.stp.sca.introspection.Resolver, Sca2Ecore {

  // debug option
  //@org.objectweb.fractal.fraclet.annotations.Attribute(value="true")
  String debug = "false";

  @Requires(name = "resolve")
  Sca2Ecore parser;

  @Requires(name = "runtime")
  RuntimeFactory runtime;

  /**
   * Logger for this class
   */

  java.util.logging.Logger log = java.util.logging.Logger
      .getLogger(ModelResolver.class.getCanonicalName());

  /**
   * Resolve missing model data
   * 
   * @param composite
   *          - The composite to add missing data
   * 
   * @return resolved SCA composite model
   */

  public Composite resolve(Composite composite) {

    // create resolver
    IncludeResolver includeResolver = new IncludeResolver(parser, runtime);

    // Use JavaInterfaceResolver from STP SCA tools
    org.eclipse.stp.sca.introspection.JavaInterfaceResolver itfresolver = new org.eclipse.stp.sca.introspection.JavaInterfaceResolver(
       runtime.getClassLoader());
    
    //JavaInterfaceResolver itfresolver = new JavaInterfaceResolver(runtime);

    CallbackInterfaceResolver callItfResolver = new CallbackInterfaceResolver(
        runtime);

    // TODO Compare with ReferenceResolver from STP SCA tools
    TargetResolver targetResolver = new TargetResolver();

    // TODO Use PromoteResolver from STP SCA tools
    // org.eclipse.stp.sca.introspection.PromoteResolver promoteResolver = new org.eclipse.stp.sca.introspection.PromoteResolver();
    
    PromoteResolver promoteResolver = new PromoteResolver();

    CompositeInterfaceResolver compositeItfResolver = new CompositeInterfaceResolver();

    // resolving on SCA model
    includeResolver.resolve(composite);
    itfresolver.resolve(composite);
    targetResolver.resolve(composite);
    promoteResolver.resolve(composite);
    callItfResolver.resolve(composite);
    compositeItfResolver.resolve(composite);

    // Validate
    try {
      Diagnostic diagnostic = Diagnostician.INSTANCE.validate(composite);

      if (diagnostic.getSeverity() != Diagnostic.OK) {
        ByteArrayOutputStream output = new ByteArrayOutputStream();
        printDiagnostic(diagnostic, "", output);
        log.warning(output.toString());
        
        // Write resulting composite file if debug is true
        if (debug.equals("true")) {
          try {
            writeComposite(composite);
          } catch (Exception e) {
            System.err.println("Error when debugging composite file");
          }
        }
        
      }
    } catch (Exception e) {
      log
          .severe("Model validation throws exception : composite definition cannot be validated");
    }

    return composite;

  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * org.ow2.frascati.factory.core.parser.Sca2Ecore#parse(java.io.InputStream)
   */

  public Composite parse(URI xmlstream) throws IOException {
    Composite composite = parser.parse(xmlstream);
    return resolve(composite);
  }

  /**
   * Print Error when validating model
   * 
   * @throws IOException
   */

  protected void printDiagnostic(Diagnostic diagnostic, String indent,
      OutputStream output) throws IOException {
    output.write(indent.getBytes());
    output.write((diagnostic.getMessage() + "\n").getBytes());
    for (Diagnostic child : diagnostic.getChildren()) {
      printDiagnostic(child, indent + "  ", output);
    }
  }

  /**
   * Write resulting composite file for debug
   * 
   * @throws FileNotFoundException
   */

  protected void writeComposite(Composite composite) throws Exception {
    Map<Object, Object> options = new HashMap<Object, Object>();

    File compositeOuput = new File("").getAbsoluteFile();
    compositeOuput = new File(compositeOuput, composite.getName()
        + ".debug.composite");
    FileOutputStream fsout = new FileOutputStream(compositeOuput);

    options.put(org.eclipse.emf.ecore.xmi.XMLResource.OPTION_ROOT_OBJECTS,
        Collections.singletonList(composite));
    ((XMLResource) composite.eResource()).save(fsout, options);
    log.warning("Write debug composite " + compositeOuput.getCanonicalPath());
  }

  public void finalize() {
    this.parser = null;
  }

}
