/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Damien Fournier
 */

package org.ow2.frascati.factory.runtime.domain;

import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.logging.Logger;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.bf.BindingFactory;
import org.objectweb.fractal.bf.BindingFactoryException;
import org.objectweb.fractal.bf.connectors.rmi.RmiConnectorConstants;
import org.objectweb.fractal.fraclet.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotations.Controller;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Policy;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.frascati.factory.core.dispatcher.Manager;
import org.ow2.frascati.factory.core.domain.DomainException;
import org.ow2.frascati.factory.core.domain.api.CompositeDomain;
import org.ow2.frascati.factory.core.instance.binding.ScaBindingProtocol;
import org.ow2.frascati.factory.core.instance.runtime.RuntimeFactory;
import org.ow2.frascati.factory.core.instance.util.ScaCompositeHelper;
import org.ow2.frascati.factory.generate.runtime.RuntimeGenerate;
import org.ow2.frascati.factory.runtime.domain.api.Domain;
import org.ow2.frascati.factory.runtime.domain.api.DomainConfig;

/**
 * FrontEnd component for FraSCAti domain management Use FraSCAti for building
 * SCA component instance, Then store them into Composite Domain. Offers method
 * for managing component of the domain
 * 
 * @author <a href="mailto:damien.fournier@inria.fr">Damien Fournier</a>
 * @version $Revision$
 * @since 0.6
 */

@org.objectweb.fractal.fraclet.annotations.Component
public class DomainManager implements Domain, DomainConfig {
  
 
  /** Logger for this class */
  protected static Logger log = Logger.getLogger(DomainManager.class.getCanonicalName());

  /** reference on the component interface of the Domain Manager */
  @Controller
  private org.objectweb.fractal.api.Component domainManager;

  //----------------------//
  // Required interfaces  //
  //----------------------//
  
  /** reference on composite domain interface */
  @Requires(name = "domain-management")
  private CompositeDomain domain;

  /** reference on load interface exported by frascati */
  @Requires(name = "load")
  private Manager load;

  /** reference on load interface exported by frascati */
  @Requires(name = "juliac")
  private RuntimeGenerate generate;

  /** reference on runtime interface exported by frascati */
  @Requires(name = "runtime")
  private RuntimeFactory runtime;

  /** reference on binding interface exported by frascati */
  @Requires(name = "binding")
  private BindingFactory binding;
  
  //-----------------------------//
  // Attributes for remote acces //
  //-----------------------------//
  
  /** service name attribute used to provide remote access to frascati domain */
  @Attribute(value=Domain.DEFAULT_DOMAIN_SERVICE_NAME)
  public String rmiServiceName;
  
  /** host name attribute used to provide remote access to frascati domain */
  @Attribute(value=Domain.DEFAULT_DOMAIN_HOST)
  public String host;
  
  /** port attribute used to provide remote access to frascati domain */
  @Attribute(value=Domain.DEFAULT_DOMAIN_PORT)
  public String port;
  
  
  /** 
   * Export the Domain Manager using the Binding Factory 
   * 
   * @throws DomainException 
   * */
  
  // Cannot export Domain and DomainConfig interface since Component is not serializable
  // TODO Remotable management interface
  //@Lifecycle(step = Step.START, policy = Policy.ASYNCHRONOUS)
  private void ExportDomain() throws DomainException {
    
    try {
      log.info("Starting new FraSCAti Domain on " + InetAddress.getLocalHost().getHostName());
    } catch (UnknownHostException e1) {
      log.info("Starting new FraSCAti Domain");
    }
    
    HashMap<String, Object> hints = new HashMap<String, Object>();
    
    // set used binding protocol
    hints.put(ScaBindingProtocol.PLUGIN_ID, "rmi");
    // retrieve classloader used by runtime component
    hints.put(ScaBindingProtocol.CLASSLOADER, runtime.getClassLoader());
    // set protocol specific parameters
    hints.put(RmiConnectorConstants.SERVICE_NAME, rmiServiceName);
    hints.put(RmiConnectorConstants.JAVA_RMI_REGISTRY_PORT, port);
    hints.put(RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS, host);
    
    try {
      log.info("Export FraSCAti domain using RMI with parameters host="+host+" port="+port+" service="+rmiServiceName);
      binding.export(domainManager, "domain-frascati", hints);
    } catch (BindingFactoryException e) {;
      throw new DomainException("Fails to export Domain for remote management",e);
    }
    
// TODO Issue when exporting both Domain and DomainConfig interfaces when
//    // Setting new rmi service name value used for domain-config interface
//    hints.put(RmiConnectorConstants.SERVICE_NAME, rmiServiceName+"-config");
//    
//    try {
//      log.info("Export FraSCAti domain config using RMI with parameters host="+host+" port="+port+" service="+rmiServiceName+"-config");
//      binding.export(domainManager, "domain-config", hints);
//    } catch (BindingFactoryException e) {;
//      throw new DomainException("Fails to export Domain Config for remote management",e);
//    }
    
  }
    

  /*
   * (non-Javadoc)
   * 
   * @see
   * org.ow2.frascati.factory.runtime.domain.api.Domain#addLibraries(java.net
   * .URL[])
   */

  public void addLibraries(URL... libs) {
    String libpath[] = new String[libs.length];
    for (int i = 0; i < libpath.length; i++) {
      libpath[i] = libs[i].getPath();
    }

    // add libraries to current classloader
    load.loadJars(libs);
    // add libraries to path
    generate.addJavaLib(libpath);
  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * org.ow2.frascati.factory.runtime.domain.api.Domain#getComposite(java.lang
   * .String)
   */

  public Component getComposite(String composite) throws DomainException {
    return getComposite(composite, null);
  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * org.ow2.frascati.factory.runtime.domain.api.Domain#getComposite(java.lang
   * .String, java.net.URL[])
   */

  public Component getComposite(String composite, URL[] libs) throws DomainException {

    Component scaComposite = domain.get(composite);

    if (scaComposite == null) {

      // Check if libraries have to be loaded
      if (libs != null)
        // add libraries if needed;
        addLibraries(libs);
      
      // Load SCA composite file/resource
      scaComposite = (Component) load.loadComposite(composite);
      
      // Start composite
      ScaCompositeHelper.start(scaComposite);
      
      // Add composite to this domain
      domain.add(scaComposite);
      
    }

    return scaComposite;
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.ow2.frascati.factory.runtime.domain.api.Domain#getComposites()
   */

  public Component[] getComposites() {
    return domain.getAll();
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.ow2.frascati.factory.runtime.domain.api.Domain#list()
   */

  public String[] getCompositesNames() {
    return domain.list();
  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * org.ow2.frascati.factory.runtime.domain.api.DomainConfig#getClassLoader()
   */

  public ClassLoader getClassLoader() {
    return runtime.getClassLoader();
  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * org.ow2.frascati.factory.runtime.domain.api.DomainConfig#setClassLoader
   * (java.net.URLClassLoader)
   */

  public void setClassLoader(URLClassLoader classloader) {
    runtime.setClassLoader(classloader);
  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * org.ow2.frascati.factory.runtime.domain.api.DomainConfig#setOutput(java
   * .io.File)
   */

  public void setOutput(File output) throws IOException {
    if (output.isDirectory()) {
      // create directory if not exist
      output.createNewFile();
      // set frascati output directory for generated files
      generate.setOutputDir(output);
    } else
      throw new IOException("Output must be a directory");
  }

  /* (non-Javadoc)
   * @see org.ow2.frascati.factory.runtime.domain.api.Domain#removeComposite(java.lang.String)
   */
  
  public void removeComposite(String name) throws DomainException {
    domain.remove(name);
  }


}
