/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.juliac;

import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.proxy.AbstractInterceptorSourceCodeGenerator;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;
import org.objectweb.fractal.juliac.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.MethodSourceCodeVisitor;
import org.ow2.frascati.tinfi.TinfiComponentInterceptor;
import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.control.component.ReconfigurableComponentContext;
import org.ow2.frascati.tinfi.control.intent.IntentHandler;
import org.ow2.frascati.tinfi.control.intent.IntentJoinPoint;
import org.ow2.frascati.tinfi.control.intent.IntentJoinPointImpl;

/**
 * This class generates the source file of Tinfi interceptors which implement
 * the logic related to the management of intents associated with components.
 * This interceptor invokes the SCA intent controller.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 0.3
 */
public class SCAIntentInterceptorSourceCodeGenerator
extends AbstractInterceptorSourceCodeGenerator {
        
    public SCAIntentInterceptorSourceCodeGenerator() {
        super();
    }

    public SCAIntentInterceptorSourceCodeGenerator(
        InterfaceType it, MembraneDesc<?> membraneDesc, boolean mergeable,
        Juliac jc ) {
        
        super(it,membraneDesc,mergeable,jc);
    }
    
    /**
     * Return the name of the suffix which is appended to generated class names.
     */
    public String getClassNameSuffix() {
        return "Intent";
    }
        
    /**
     * Return <code>true</code> if some code must be generated by the current
     * source code generator.
     */
    public boolean match() {
        // Match all interfaces, including control ones
        return true;
    }

    @Override
    public void setMergeable( boolean mergeable ) {
        if(mergeable) {
            String msg = "This generator can not generate mergeable code";
            throw new IllegalArgumentException(msg);
        }
    }

    @Override
    public String[] getImplementedInterfaceNames() {
        return new String[]{it.getFcItfSignature()};
    }
    
    @Override
    public String getSuperClassName() {
        return TinfiComponentInterceptor.class.getName();
    }

    
    @Override
    public void generateStaticParts( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        
        /*
         * Initialize the METHODS field (declared below in method
         * generateFields).
         */
        BlockSourceCodeVisitor bv = cv.visitStaticPart();
        bv.visitBegin();
        bv.visitln("    try {");
        bv.visitln("      METHODS = new java.lang.reflect.Method[]{");
        
        // Iterate on methods
        UnifiedMethod[] proxymethods = proxycl.getMethods();
        String proxyclname = proxycl.getName();
        for (int i = 0; i < proxymethods.length; i++) {
            
            UnifiedMethod proxym = proxymethods[i];
            String proxymname = proxym.getName();
            UnifiedClass[] paramtypes = proxym.getParameterTypes();

            bv.visit("        ");
            bv.visit(proxyclname);
            bv.visit(".class.getMethod(\"");
            bv.visit(proxymname);
            bv.visit("\"");
            for (UnifiedClass paramtype : paramtypes) {
                String paramtypename = paramtype.getName();
                bv.visit(",");
                bv.visit(paramtypename);
                bv.visit(".class");
            }
            bv.visitln("),");
        }
        bv.visitln("      };");
        
        bv.visitln("    }");
        bv.visitln("    catch( NoSuchMethodException e ) {");
        bv.visit  ("      throw new ");
        bv.visit  (TinfiRuntimeException.class.getName());
        bv.visitln("(e);");
        bv.visitln("    }");
        bv.visitEnd();        
    }

    @Override
    public void generateFields( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        
        /*
         * Generate a METHODS field to store the array of component methods
         * generated by the interceptor.
         */
        int modifiers = Modifier.PRIVATE + Modifier.STATIC;
        String type = "java.lang.reflect.Method[]";
        String name = "METHODS";
        cv.visitField( modifiers, type, name, null );
        
        /*
         * Generate a subclass of IntentJoinPointImpl per method defined in
         * proxycl. Instances of the generated class reifies invocations on the
         * associated method.
         */
        UnifiedMethod[] proxymethods = proxycl.getMethods();
        String proxyclname = proxycl.getName();
        for (int i = 0; i <proxymethods.length; i++) {
            
            UnifiedMethod proxym = proxymethods[i];
            String proxymname = proxym.getName();
            String proxymrettypename = proxym.getReturnType().getName();
            UnifiedClass[] paramtypes = proxym.getParameterTypes();
            
            // Inner class header
            ClassSourceCodeVisitor innercv = cv.visitInnerClass();
            int innermodifiers = Modifier.PRIVATE + Modifier.STATIC;
            String innername = "IntentJoinPointImplForMethod"+i;
            innercv.visit(
                innermodifiers, innername, null,
                IntentJoinPointImpl.class.getName(), null );
            
            // Constructor
            String[] args = new String[ paramtypes.length + 5];
            args[0] = "java.util.List<"+IntentHandler.class.getName()+"> handlers";
            args[1] = ReconfigurableComponentContext.class.getName()+" compctx";
            args[2] = Interface.class.getName()+" itf";
            args[3] = proxyclname+" intentTarget";
            args[4] = "java.lang.reflect.Method intentMethod";
            for (int j = 0; j < paramtypes.length; j++) {
                String paramtypename = paramtypes[j].getName();
                args[j+5] = paramtypename+" arg"+j;
            }
            
            // Constructor body
            MethodSourceCodeVisitor mv =
                innercv.visitConstructor( Modifier.PUBLIC, null, args, null );
            mv.visitBegin();
            mv.visit("    super(handlers,compctx,itf,intentTarget,intentMethod");
            for (int j = 0; j < paramtypes.length; j++) {
                mv.visit(",arg"+j);
            }
            mv.visitln(");");
            mv.visitEnd();
            
            // proceed() method
            mv = innercv.visitMethod(
                Modifier.PUBLIC, null, "Object", "proceed", null,
                new String[]{"Throwable"});
            mv.visitBegin();
            mv.visitln("    if( index == handlers.size() ) {");
            mv.visit  ("      ");
            if( ! proxymrettypename.equals("void") ) {
                mv.visit("return ");
            }
            mv.visit  ("base((");
            mv.visit  (proxyclname);
            mv.visit  (")intentTarget");
            for (int j = 0; j < paramtypes.length; j++) {
                UnifiedClass paramtype = paramtypes[j];
                if( paramtypes[j].isPrimitive() ) {
                    paramtype = paramtype.box();
                }
                String paramtypename = paramtype.getName();
                mv.visit(",");
                if( ! paramtypename.equals("java.lang.Object") ) {
                    mv.visit("(");
                    mv.visit(paramtypename);
                    mv.visit(")");
                }
                mv.visit("intentMethodArguments[");
                mv.visit(Integer.toString(j));
                mv.visit("]");
            }
            mv.visitln(");");
            if( proxymrettypename.equals("void") ) {
                mv.visit("      return null;");
            }
            mv.visitln("    }");
            mv.visitln("    else {");
            mv.visitln("      return super.proceed();");
            mv.visitln("    }");
            
            // End of the proceed() method
            mv.visitEnd();
            
            /*
             * base() method.
             * 
             * In theory, the base() method is not strictly needed. One should
             * be able to invoke the base method from the proceed method
             * generated just before. It happens that the JDT compiler can not
             * decide which method to invoke in the case of the ComponentContext
             * interface with the two cast(Object) methods. The fact of
             * introducing the base() method solves the issue.
             */
            String[] genericReturnType = proxym.getTypeParameters();
            String returnType = proxym.getGenericReturnType();
            String[] params = proxym.getParameters();
            List<String> parameters = new ArrayList<String>();
            parameters.addAll( Arrays.asList(params) );
            // add(int,String) is not supported by the type returned by Arrays.asList
            parameters.add(0,proxyclname+" base");
            String[] exceptions = proxym.getGenericExceptionTypes();
            mv = innercv.visitMethod(
                Modifier.PRIVATE, genericReturnType, returnType, "base",
                parameters.toArray(new String[parameters.size()]), exceptions );
            mv.visitBegin();
            
            // base() method body
            UnifiedClass[] ptypes = proxym.getParameterTypes();
            String rtypename = proxym.getGenericReturnType();

            mv.visit("    ");
            if( ! rtypename.equals("void") ) {
                mv.visit(rtypename+" ret = ");
            }
            
            mv.visit("base.");
            mv.visit(proxymname);
            mv.visit("(");
            for (int j = 0; j < ptypes.length; j++) {
                if(j!=0)  mv.visit(",");
                mv.visit("arg");
                mv.visit(Integer.toString(j));
            }
            mv.visitln(");");
            if( ! rtypename.equals("void") ) {
                mv.visitln("    return ret;");
            }
            
            // End of the base() method
            mv.visitEnd();
            
            // End of the inner class
            innercv.visitEnd();
        }
    }

    @Override
    public void generateMethodInitFcController( MethodSourceCodeVisitor mv, UnifiedClass proxycl ) {
        mv.visitln("    initIntentHandlersMap(METHODS);");
        mv.visitln("    super.initFcController(ic);");
    }
    
    @Override
    public void generateMethodClone( MethodSourceCodeVisitor mv, UnifiedClass proxycl ) {        
        mv.visitln("    initFcClone(clone);");
    }
    
    @Override
    public void generateProxyMethodBodyBeforeCode(
        MethodSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        /*
         * Retrieve the index of proxym in the array of methods for proxycl.
         */
        int idx = -1;
        String proxymstring = proxym.toString();
        UnifiedMethod[] ums = proxycl.getMethods();
        for (int i = 0; i < ums.length; i++) {
            UnifiedMethod um = ums[i];
            String umstring = um.toString();
            /*
             * TODO fix issue with the 2 cast() methods in ComponentContext
             * which have the same signature when omitting generics cast(Object)
             * <B, R extends CallableReference<B>> R cast(B target)
             * <B> ServiceReference<B> cast(B target) 
             */
            if( umstring.equals(proxymstring) ) {
                idx = i;
                break;
            }
        }
        if( idx == -1 ) {
            String msg =
                proxym+" should have been found in the array of methods of "+
                proxycl;
            throw new TinfiRuntimeException(msg);
        }
        
        // Retrieve the handlers associated with the component
        mv.visit  ("    java.util.List<");
        mv.visit  (IntentHandler.class.getName());
        mv.visit  ("> handlers = intentHandlersMap.get(METHODS[");
        mv.visit  (Integer.toString(idx));
        mv.visitln("]);");
        
        
        // Retrieve the component context
        mv.visit  ("    ");
        mv.visit  (ReconfigurableComponentContext.class.getName());
        mv.visitln(" compctx = getFcCompCtxCtrlItf();");
        
        // Retrieve the interface associated with the interceptor
        mv.visit  ("    ");
        mv.visit  (Interface.class.getName());
        mv.visitln(" itf = getFcItf();");
        
        // Begin of try/finally block
        mv.visitln("    try {");

        /*
         * If the handlers list is empty don't create the IntentJoinPoint
         * instance and directly call the delegate.
         */
        String proxymname = proxym.getName();
        UnifiedClass[] ptypes = proxym.getParameterTypes();
        String rtypename = proxym.getGenericReturnType();
        
        mv.visitln("      if( handlers.size() == 0 ) {");
        
        // base call
        mv.visit("        ");
        if( ! rtypename.equals("void") ) {
            mv.visit(rtypename+" ret = ");
        }
        
        mv.visit("_impl.");
        mv.visit(proxymname);
        mv.visit("(");
        for (int j = 0; j < ptypes.length; j++) {
            if(j!=0)  mv.visit(",");
            mv.visit("arg");
            mv.visit(Integer.toString(j));
        }
        mv.visitln(");");
        if( ! rtypename.equals("void") ) {
            mv.visitln("        return ret;");
        }
        mv.visitln("      }");
        
        // Intent join point creation and invocation
        mv.visitln("      else {");
        mv.visit  ("        ");
        mv.visit  (IntentJoinPoint.class.getName());
        mv.visitln(" ijp = ");
        mv.visit  ("          new IntentJoinPointImplForMethod");
        mv.visit  (Integer.toString(idx));
        mv.visit  ("(handlers,compctx,itf,_impl,METHODS[");
        mv.visit  (Integer.toString(idx));
        mv.visit  ("]");
        for (int i = 0; i < ptypes.length; i++) {
            mv.visit(",arg");
            mv.visit(new Integer(i).toString());
        }
        mv.visitln(");");        
    }
    
    @Override
    public void generateProxyMethodBodyFinallyCode(
        MethodSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        // End of else block
        mv.visitln("      }");
        
        // End of try/finally block
        mv.visitln("    }");
        mv.visitln("    catch( Throwable t ) {");
        
        // Handle exception thrown by the method
        UnifiedClass[] excepts = proxym.getExceptionTypes();
        for (UnifiedClass except : excepts) {
            mv.visit  ("      if( t instanceof ");
            mv.visit  (except.getName());
            mv.visitln(" ) {");
            mv.visit  ("        throw (");
            mv.visit  (except.getName());
            mv.visitln(") t;");
            mv.visitln("      }");
        }
        
        // Handle RuntimeExceptions
        mv.visitln("      if( t instanceof RuntimeException ) {");
        mv.visitln("        throw (RuntimeException) t;");
        mv.visitln("      }");
        
        // Handle other types of exception
        mv.visit  ("      throw new ");
        mv.visit  (TinfiRuntimeException.class.getName());
        mv.visitln("(t);");
        mv.visitln("    }");
    }
}
