/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.juliac;

import java.lang.reflect.Modifier;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.ComponentInterface;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.ClassGenerator;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.proxy.AbstractProxyClassGenerator;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;
import org.objectweb.fractal.juliac.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.MethodSourceCodeVisitor;
import org.osoa.sca.CallableReference;
import org.osoa.sca.ConversationEndedException;
import org.osoa.sca.ServiceReference;
import org.osoa.sca.annotations.Callback;
import org.osoa.sca.annotations.Conversational;
import org.osoa.sca.annotations.EndsConversation;
import org.osoa.sca.annotations.OneWay;
import org.ow2.frascati.tinfi.CallbackManager;
import org.ow2.frascati.tinfi.ConversationManager;
import org.ow2.frascati.tinfi.osoa.ConversationImpl;
import org.ow2.frascati.tinfi.osoa.ConversationState;
import org.ow2.frascati.tinfi.osoa.ServiceReferenceImpl;

/**
 * This class generates the source code of SCA service reference
 * implementations.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class ServiceReferenceClassGenerator
extends AbstractProxyClassGenerator {
    
    public ServiceReferenceClassGenerator() {
        super();
    }
    
    public ServiceReferenceClassGenerator(
        InterfaceType it, MembraneDesc<?> membraneDesc, boolean mergeable,
        Juliac jc ) {
        
        super(it,membraneDesc,mergeable,jc);
    }
    
    /**
     * Return <code>true</code> if some code must be generated by this
     * generator.
     */
    public boolean match() {
        return true;
    }

    /**
     * Return the name of the suffix which is appended to generated class names.
     */
    public String getClassNameSuffix() {
        return "FcSR";
    }

    @Override
    public void setMergeable( boolean mergeable ) {
        if(mergeable) {
            String msg = "This generator can not generate mergeable code";
            throw new IllegalArgumentException(msg);
        }
    }

    @Override
    public String[] getClassGenericTypeParameters() {
        // TODO handle generation for generic interfaces (e.g. o.o.f.bf.BindingFactoryPlugin)
        return new String[]{"B extends "+it.getFcItfSignature()};
    }

    @Override
    public String getSuperClassName() {
        return ServiceReferenceImpl.class.getName()+"<B>";
    }

    @Override
    public String[] getImplementedInterfaceNames() {
        return new String[]{it.getFcItfSignature()};
    }
    
    @Override
    public void generateConstructors(
        ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        
        super.generateConstructors(cv,proxycl);
        
        // public constructor(Class<B>,B)
        String[] parameters =
            new String[]{ "Class<B> businessInterface", "B service" };
        MethodSourceCodeVisitor mv =
            cv.visitConstructor(Modifier.PUBLIC,null,parameters,null);
        mv.visitBegin();
        mv.visitln("    super(businessInterface,service);");
        mv.visitEnd();
    }
    
    @Override
    public void generateProxyMethodBodyBeforeCode(
        MethodSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        super.generateProxyMethodBodyBeforeCode(mv,proxycl,proxym);
    
        conversationalItf = proxycl.getAnnotation(Conversational.class);
        oneWayMeth = proxym.getAnnotation(OneWay.class);
        
        callbackClass = Void.class;   // callback.value() defaults to Void
        Callback callback = proxycl.getAnnotation(Callback.class);
        if( callback != null ) {
            callbackClass = callback.value();
        }

        // Retrieve the delegate
        mv.visitln("    final B impl = getService();");
        mv.visitln("    final "+ServiceReference.class.getName()+"<B> sr = this;");
        
        // OneWay method
        if( oneWayMeth != null ) {
            mv.visitln("    new Thread() {");
            mv.visitln("      public void run() {");
        }

        // Method defined in a conversational interface
        if( conversationalItf != null ) {
            mv.visitln("        "+ConversationImpl.class.getName()+" conv = getConversation();");
            mv.visitln("        "+ConversationState.class.getName()+" convstate = conv.getFcState();");
            mv.visitln("        if( convstate.equals("+ConversationState.class.getName()+".STOPPED) ) {");
            mv.visitln("          throw new "+ConversationEndedException.class.getName()+"();");
            mv.visitln("        }");
            mv.visitln("        if( convstate.equals("+ConversationState.class.getName()+".CREATED) ) {");
            mv.visitln("          String conversationId = Thread.currentThread().toString()+':'+getService().toString();");
            mv.visitln("          conv.setConversationID(conversationId);");
            mv.visitln("          conv.start();");
            mv.visitln("        }");
            mv.visit  ("        java.util.Stack<"+ConversationImpl.class.getName()+"> convstack = ");
            mv.visitln(ConversationManager.class.getName()+".get();");
            mv.visitln("        convstack.push(conv);");
        }
        
        // Method defined in an interface with a callback
        if( callbackClass != Void.class ) {
            
            InterfaceType cbit =
                new BasicInterfaceType(
                    "callback",callbackClass.getName(),false,false,false);
            ClassGenerator cg =
                new CallBackInterfaceClassGenerator(cbit,null,mergeable,jc);
            String cbitfclname = cg.getTargetClassName();
            
            cg = new CallableReferenceClassGenerator(cbit,null,mergeable,jc);
            String cbcrclname = cg.getTargetClassName();
            
            mv.visit  ("        "+ComponentInterface.class.getName()+" ci = (");
            mv.visitln(ComponentInterface.class.getName()+") getService();");
            mv.visitln("        "+Component.class.getName()+" owner = ci.getFcItfOwner();");
            mv.visit  ("        "+InterfaceType.class.getName()+" it = ");
            mv.visit  (Utils.javaify(cbit).toString());
            mv.visitln(";");
            mv.visitln("        "+cbitfclname+" cbci = new "+cbitfclname+"(owner,\"callback\",it,false,null);");
            
            mv.visit  ("        "+CallableReference.class.getName()+"<"+callbackClass.getName()+"> cb = ");
            mv.visit  ("new "+cbcrclname+"<"+callbackClass.getName()+">(");
            mv.visitln(callbackClass.getName()+".class,cbci);");
            mv.visitln("        sr.setCallback(cb);");
            mv.visit  ("        java.util.Stack<"+CallableReference.class.getName()+"<?>> cbstack = ");
            mv.visitln(CallbackManager.class.getName()+".get();");
            mv.visitln("        cbstack.push(cb);");
        }
    }
    private Conversational conversationalItf;
    private OneWay oneWayMeth;
    private Class<?> callbackClass;
    
    @Override
    public void generateProxyMethodBodyAfterCode(
        MethodSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        super.generateProxyMethodBodyAfterCode(mv,proxycl,proxym);
        
        EndsConversation end = proxym.getAnnotation(EndsConversation.class);
        if( end != null ) {
            mv.visitln("    conv.end();");
        }
        if( conversationalItf != null ) {
            mv.visitln("    conv.access();");
            mv.visitln("    convstack.pop();");
        }
        if( callbackClass != Void.class ) {
            mv.visitln("    cbstack.pop();");
            mv.visitln("    sr.setCallback(null);");
        }
        if( oneWayMeth != null ) {
            mv.visitln("      }");
            mv.visitln("    }.start();");
        }
    }

    public String getDelegatingInstance(
        UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        return "impl";
    }
}
