/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.metadata;

import java.lang.annotation.Annotation;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * <p>
 * This class provides a way for dynamically attaching and retrieving metadata
 * associated with a class. Metadata can be attached to the class itself or to
 * its members (constructors, fields, methods). Metadata consists of:
 * </p>
 * 
 * <ul>
 * <li>either raw data which can be set and get by the API provided by this class,</li>
 * <li>or Java 5 annotations with the runtime retention policy.</li>
 * </ul>
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 0.4.3
 */
public class MetaData<T> {
    
    /**
     * @param cl  the class supporting the metadata
     */
    public MetaData( Class<T> cl ) {
        this.cl = cl;
    }
    
    /** The class supporting the metadata. */
    private Class<T> cl;
    
    /**
     * Return the class supporting the metadata.
     */
    public Class<T> getSupportClass() {
        return cl;
    }
    
    @Override
    public String toString() {
        return cl.getName();
    }
    
    
    // -----------------------------------------------------------------------
    // API for retrieving metadata associated with the support class
    // -----------------------------------------------------------------------
    
    /**
     * If the support class is associated with metadata of the specified type,
     * return this metadata. Else return <code>null</code>.
     */
    public <A extends Annotation> A getAnnotation( Class<A> annotcl ) {

        // First, search in dynamically set metadata
        List<Object> mds = getMetaData();
        for (Object md : mds) {
            if( annotcl.isAssignableFrom(md.getClass()) ) {
                @SuppressWarnings("unchecked")
                A annot2 = (A) md;
                return annot2;
            }
        }
        
        // Second, search in Java 5 annotations
        A annot = cl.getAnnotation(annotcl);
        if( annot != null ) {
            return annot;
        }
        
        // No match
        return null;
    }
    
    /**
     * Return the metadata associated with the support class.
     */
    public Object[] getAnnotations() {
        
        List<Object> mds = getMetaData();
        mds = new ArrayList<Object>(mds);  // Work on a copy

        Annotation[] annots = cl.getAnnotations();
        for (Annotation annot : annots) {
            Class<?> annotcl = annot.getClass();
            /*
             * Iterate on dynamically set metadata to check whether a
             * corresponding annotation exists.
             */
            boolean exists = false;
            for (Object md : mds) {
                Class<?> cl = md.getClass();
                if( cl.equals(annotcl) ) {
                    exists = true;
                }
            }
            if(!exists) {
                /*
                 * Only consider the annotation if it has not been set in the
                 * metadata. Metadata has a priority level higher than that of
                 * annotations in order to be able to override annotations with
                 * some dynamically set metadata.
                 */
                mds.add(annot);
            }
        }
        
        Object[] all = mds.toArray();
        return all;
    }
    
    /**
     * <p>
     * Return all fields associated with some metadata of the specified
     * annotation type. The query is performed on:
     * </p>
     * 
     * <ul>
     * <li>Java 5 annotated fields declared in the support class,</li>
     * <li>metadata associated fields declared in the support class,</li>
     * <li>Java 5 annotated public fields inherited by the support class.</li>
     * 
     * @param annotcl  the annotation type
     * @return         the corresponding array of fields
     */
    public Field[] getAllAnnotatedFields( Class<? extends Annotation> annotcl ) {
        
        List<Field> result = new ArrayList<Field>();
        
        // First, search in fields declared in the class
        Field[] fields = cl.getDeclaredFields();
        for (Field field : fields) {
            if( field.getAnnotation(annotcl) != null ) {
                result.add(field);
            }
            else {
                // Search in dynamically set metadata
                List<Object> mds = getMetaData(field);
                if( mds != null ) {
                    for (Object md : mds) {
                        if( annotcl.isAssignableFrom(md.getClass()) ) {
                            result.add(field);
                        }
                    }
                }
            }
        }
        
        // Second, search in declared and inherited public methods
        fields = cl.getFields();
        for (Field field : fields) {
            // Skip public fields declared in the class
            if( ! result.contains(field) ) {
                if( field.getAnnotation(annotcl) != null ) {
                    result.add(field);
                }
            }
        }
        
        return result.toArray(new Field[result.size()]);
    }

    /**
     * <p>
     * Return all methods associated with some metadata of the specified
     * annotation type. The query is performed on:
     * </p>
     * 
     * <ul>
     * <li>Java 5 annotated methods declared in the support class,</li>
     * <li>metadata associated methods declared in the support class,</li>
     * <li>Java 5 annotated public methods inherited by the support class.</li>
     * 
     * @param annotcl  the annotation type
     * @return         the corresponding array of methods
     */
    public Method[] getAllAnnotatedMethods( Class<? extends Annotation> annotcl ) {
        
        List<Method> result = new ArrayList<Method>();
        
        // First, search in methods declared in the class
        Method[] methods = cl.getDeclaredMethods();
        for (Method method : methods) {
            if( method.getAnnotation(annotcl) != null ) {
                result.add(method);
            }
            else {
                // Search in dynamically set metadata
                List<Object> mds = getMetaData(method);
                if( mds != null ) {
                    for (Object md : mds) {
                        if( annotcl.isAssignableFrom(md.getClass()) ) {
                            result.add(method);
                        }
                    }
                }
            }
        }
        
        // Second, search in declared and inherited public methods
        methods = cl.getMethods();
        for (Method method : methods) {
            // Skip public methods declared in the class
            if( ! result.contains(method) ) {
                if( method.getAnnotation(annotcl) != null ) {
                    result.add(method);
                }
            }
        }
        
        return result.toArray(new Method[result.size()]);
    }

    /**
     * <p>
     * Return all methods and fields associated with some metadata of the
     * specified annotation type. The query is performed on:
     * </p>
     * 
     * <ul>
     * <li>Java 5 annotated methods and fields declared in the support class,</li>
     * <li>metadata associated methods and fields declared in the support class,</li>
     * <li>Java 5 annotated public methods and fields inherited by the support class.</li>
     * 
     * @param annotcl  the annotation type
     * @return         the corresponding array of methods and fields
     */
    public AccessibleObject[] getAllAnnotatedMethodsAndFields(
        Class<? extends Annotation> annotcl ) {
        
        Method[] methods = getAllAnnotatedMethods(annotcl);
        Field[] fields = getAllAnnotatedFields(annotcl);
        
        AccessibleObject[] result = new AccessibleObject[ methods.length + fields.length ];
        System.arraycopy( methods, 0, result, 0, methods.length );
        System.arraycopy( fields, 0, result, methods.length, fields.length );
        
        return result;
    }

    /**
     * <p>
     * Return all unannotated setter methods. The query is performed on:
     * </p>
     * 
     * <ul>
     * <li>methods declared in the support class,</li>
     * <li>public methods inherited by the support class.</li>
     * </ul>
     * 
     * @return  the corresponding map of methods indexed by property names
     *          (the property set by the method).
     * @since 0.4.4
     */
    public Map<String,Method> getAllUnAnnotatedSetterMethods() {
        
        Map<String,Method> result = new HashMap<String,Method>();
        
        // First, search in methods declared in the class
        Method[] methods = cl.getDeclaredMethods();
        for (Method method : methods) {
            if( Util.isSetterMethod(method) ) {
                Annotation[] annots = method.getAnnotations();
                if( annots.length == 0 ) {
                    String prop = Util.getPropertyName(method);
                    if( result.containsKey(prop) ) {
                        // TODO support polymorphic setter methods when a property and a reference share the same name (see p12 of the specs for the rules)
                        String msg =
                            "More than one setter method in class "+
                            cl.getName()+" for property "+prop;
                        throw new UnsupportedOperationException(msg);
                    }
                    result.put(prop,method);
                }
            }
        }
        
        // Second, search in declared and inherited public methods
        methods = cl.getMethods();
        for (Method method : methods) {
            if( Util.isSetterMethod(method) ) {
                Annotation[] annots = method.getAnnotations();
                if( annots.length == 0 ) {
                    String prop = Util.getPropertyName(method);
                    if( result.containsKey(prop) &&
                        ! result.containsValue(method) ) {
                        // Skip public methods already added by the previous loop
                        String msg =
                            "More than one setter method in class "+
                            cl.getName()+" for property "+prop;
                        throw new UnsupportedOperationException(msg);
                    }
                    result.put(prop,method);
                }
            }
        }
        
        return result;
    }

    /**
     * <p>
     * Return all constructors declared in the support class and associated with
     * some metadata of the specified annotation type. The query is performed
     * on:
     * </p>
     * 
     * <ul>
     * <li>Java 5 annotated constructors declared in the support class,</li>
     * <li>metadata associated constructors declared in the support class.</li>
     * 
     * @param annotcl  the annotation type
     * @return         the corresponding array of constructors
     */
    public Constructor<?>[] getDeclaredAnnotatedConstructors(
        Class<? extends Annotation> annotcl ) {
        
        List<Constructor<?>> result = new ArrayList<Constructor<?>>();
        
        Constructor<?>[] constructors = cl.getDeclaredConstructors();
        for (Constructor<?> constructor : constructors) {
            if( constructor.getAnnotation(annotcl) != null ) {
                result.add(constructor);
            }
            else {
                // Search in dynamically set metadata
                List<Object> mds = getMetaData(constructor);
                if( mds != null ) {
                    for (Object md : mds) {
                        if( annotcl.isAssignableFrom(md.getClass()) ) {
                            result.add(constructor);
                        }
                    }
                }
            }
        }
        
        return result.toArray(new Constructor[result.size()]);
    }
    
    
    // -----------------------------------------------------------------------
    // API for setting and getting metadata associated with the support class
    // -----------------------------------------------------------------------
    
    /**
     * Add the specified metadata element in the list of metadata associated
     * with the support class.
     * 
     * @param md  the metadata element
     */
    public void addMetaData( Object md ) {
        clmd.add(md);
    }

    /**
     * Add the specified metadata element in the list of metadata associated
     * with the specified constructor.
     * 
     * @param c   the constructor
     * @param md  the metadata element
     */
    public void addMetaData( Constructor<?> c, Object md ) {
        checkOwnership(c);
        List<Object> mds = ctormd.get(c);
        if( mds == null ) {
            mds = new ArrayList<Object>();
            ctormd.put(c,mds);
        }
        mds.add(md);
    }
    
    /**
     * Add the specified metadata element in the list of metadata associated
     * with the specified field.
     * 
     * @param f   the field
     * @param md  the metadata element
     */
    public void addMetaData( Field f, Object md ) {
        checkOwnership(f);
        List<Object> mds = fieldmd.get(f);
        if( mds == null ) {
            mds = new ArrayList<Object>();
            fieldmd.put(f,mds);
        }
        mds.add(md);
    }
    
    /**
     * Add the specified metadata element in the list of metadata associated
     * with the specified method.
     * 
     * @param m   the method
     * @param md  the metadata element
     */
    public void addMetaData( Method m, Object md ) {
        checkOwnership(m);
        List<Object> mds = methmd.get(m);
        if( mds == null ) {
            mds = new ArrayList<Object>();
            methmd.put(m,mds);
        }
        mds.add(md);
    }

    /**
     * Return <code>true</code> if some metadata is associated with the support
     * class.
     * 
     * @return  <code>true</code> if some metadata is associated
     */
    public boolean containsMetaData() {
        boolean b = ! clmd.isEmpty();
        return b;
    }
    
    /**
     * Return <code>true</code> if some metadata is associated with the
     * specified constructor.
     * 
     * @param c  the constructor
     * @return   <code>true</code> if some metadata is associated
     */
    public boolean containsMetaData( Constructor<?> c ) {
        checkOwnership(c);
        boolean b = ctormd.containsKey(c);
        return b;
    }
    
    /**
     * Return <code>true</code> if some metadata is associated with the
     * specified field.
     * 
     * @param f  the field
     * @return   <code>true</code> if some metadata is associated
     */
    public boolean containsMetaData( Field f ) {
        checkOwnership(f);
        boolean b = fieldmd.containsKey(f);
        return b;
    }
    
    /**
     * Return <code>true</code> if some metadata is associated with the
     * specified method.
     * 
     * @param m  the method
     * @return   <code>true</code> if some metadata is associated
     */
    public boolean containsMetaData( Method m ) {
        checkOwnership(m);
        boolean b = methmd.containsKey(m);
        return b;
    }
    
    /**
     * Return the list of metadata elements associated with the support class.
     * Return an empty list if no metadata has been associated.
     * 
     * @return  the list of metadata elements
     */
    public List<Object> getMetaData() {
        return clmd;
    }
    
    /**
     * Return the list of metadata elements associated with the specified field.
     * Return <code>null</code> if no metadata has been associated.
     * 
     * @param f  the method
     * @return   the list of metadata elements
     */
    public List<Object> getMetaData( Constructor<?> c ) { 
        checkOwnership(c);
        List<Object> ret = ctormd.get(c);
        return ret;
    }
    
    /**
     * Return the list of metadata elements associated with the specified field.
     * Return <code>null</code> if no metadata has been associated.
     * 
     * @param f  the method
     * @return   the list of metadata elements
     */
    public List<Object> getMetaData( Field f ) { 
        checkOwnership(f);
        List<Object> ret = fieldmd.get(f);
        return ret;
    }
    
    /**
     * Return the list of metadata elements associated with the specified
     * method. Return <code>null</code> if no metadata has been associated.
     * 
     * @param m  the method
     * @return   the list of metadata elements
     */
    public List<Object> getMetaData( Method m ) { 
        checkOwnership(m);
        List<Object> ret = methmd.get(m);
        return ret;
    }
    
    
    // -----------------------------------------------------------------------
    // Structures holding metadata
    // -----------------------------------------------------------------------
    
    private List<Object> clmd = new ArrayList<Object>();
    private Map<Constructor<?>,List<Object>> ctormd = new HashMap<Constructor<?>,List<Object>>();
    private Map<Field,List<Object>> fieldmd = new HashMap<Field,List<Object>>();
    private Map<Method,List<Object>> methmd = new HashMap<Method,List<Object>>();

    
    // -----------------------------------------------------------------------
    // Utility methods
    // -----------------------------------------------------------------------
    
    /**
     * Check that the specified member is declared by the support class.
     * 
     * @param member  the member to check
     * @throws IllegalArgumentException
     *      if the specified member is not declared by the support class
     */
    private void checkOwnership( Member member )
    throws IllegalArgumentException {
        
        Class<?> declaringClass = member.getDeclaringClass();
        
        if( ! cl.equals(declaringClass) ) {
            String msg = member+" is not declared by "+cl.getName();
            throw new IllegalArgumentException(msg);
        }
    }
}
