/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.control.content;

import java.util.Stack;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.Controller;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.factory.ChainedInstantiationException;
import org.osoa.sca.RequestContext;
import org.osoa.sca.annotations.Destroy;
import org.ow2.frascati.tinfi.control.component.ReconfigurableComponentContext;
import org.ow2.frascati.tinfi.control.content.scope.CompositeScopeManager;
import org.ow2.frascati.tinfi.control.content.scope.ConversationScopeManager;
import org.ow2.frascati.tinfi.control.content.scope.RequestScopeManager;
import org.ow2.frascati.tinfi.control.content.scope.ScopeManager;
import org.ow2.frascati.tinfi.control.content.scope.StatelessScopeManager;
import org.ow2.frascati.tinfi.control.property.SCAPropertyController;

/**
 * Mixin layer for implementing the {@link SCAContentController} interface.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public abstract class SCAContentControllerMixin
implements Controller, SCAContentController {
    
    // -------------------------------------------------------------------------
    // Private constructor
    // -------------------------------------------------------------------------

    private SCAContentControllerMixin () {}

    
    // -------------------------------------------------------------------------
    // Implementation of the Controller interface
    // -------------------------------------------------------------------------

    /**
     * The metadata for the content class associated to the component controlled
     * by the current controller.
     */
    private ContentClassMetaData ccmd;
    
    /**
     * The scope manager for the component controlled by the current controller.
     */
    private ScopeManager sm;

    /**
     * Initializes the fields of this mixin and then calls the overriden method.
     *
     * @param ic information about the component to which this controller object
     *      belongs.
     * @throws InstantiationException if the initialization fails.
     */
    public void initFcController( final InitializationContext ic )
    throws InstantiationException {
        
        /*
         * Propagate the invocation to other mixin layers.
         */
        _super_initFcController(ic);
        
        /*
         * Initialize this layer.
         * Do it after the propagation in order for _this_XXX fields to be
         * initialized.
         */
        String fcContentDesc = (String) ic.content;
        innerSetFcContentClassName(fcContentDesc);
    }
  
    /**
     * Declares the name of the content class which should be used. This
     * operation can only be performed if the component is stopped. This method
     * is synchronized to prevent inconsistent concurrent changes.
     * 
     * @since 0.6
     */
    private synchronized void innerSetFcContentClassName( String name )
    throws InstantiationException {
        
        /*
         * Content class meta data and scope manager retrievals.
         */
        try {
            ccmd = ContentClassMetaData.get(name);
        }
        catch( IllegalContentClassMetaData e ) {
            throw new ChainedInstantiationException(
                e,_this_weaveableC,e.getMessage());
        }
        
        if( ccmd.scope == null ||
            ccmd.scope.value().toUpperCase().equals("STATELESS") ) {
            
            sm = new StatelessScopeManager(
                _this_weaveableC, _this_weavablePC, _this_weavableCompCtx,
                _this_weaveableOptNC, ccmd );
        }
        else if( ccmd.scope.value().toUpperCase().equals("REQUEST") ) {
            sm = new RequestScopeManager(
                _this_weaveableC, _this_weavablePC, _this_weavableCompCtx,
                _this_weaveableOptNC, ccmd );
        }
        else if( ccmd.scope.value().toUpperCase().equals("COMPOSITE") ) {
            sm = new CompositeScopeManager(
                _this_weaveableC, _this_weavablePC, _this_weavableCompCtx,
                _this_weaveableOptNC, ccmd );
        }
        else if( ccmd.scope.value().toUpperCase().equals("CONVERSATION") ) {
            sm = new ConversationScopeManager(
                _this_weaveableC, _this_weavablePC, _this_weavableCompCtx,
                _this_weaveableOptNC, ccmd );
        }
        else {
            String msg = "Unsupported scope: "+ccmd.scope.value();
            throw new InstantiationException(msg);
        }
    }
    
    
    // -------------------------------------------------------------------------
    // Implementation of the SCAContentController interface
    // -------------------------------------------------------------------------
    
    /**
     * @since 0.6
     */
    public void setFcContentClassName( String name )
    throws IllegalLifeCycleException, InstantiationException {
        
        /*
         * Lifecycle check.
         */
        String state = _this_weaveableOptLC.getFcState();
        if( ! state.equals(LifeCycleController.STOPPED) ) {
            String msg = "Component should be stopped";
            throw new IllegalLifeCycleException(msg);
        }
        
        /*
         * Content class meta data and scope manager retrievals.
         */
        innerSetFcContentClassName(name);
    }
    
    public void eagerInit() throws InstantiationException {
        if( ccmd.eagerinit != null ) {
            if( !(sm instanceof CompositeScopeManager) ) {
                String msg =
                    "Eager initialization is meaningess in scopes other than "+
                    "COMPOSITE (scope is: "+ccmd.scope.value();
                throw new InstantiationException(msg);
            }
            sm.getFcContent();
        }
    }

    public Stack<RequestContext> getRequestContextStack() {
        return sm.getRequestContextStack();
    }
    
    public void setRequestContextStack( Stack<RequestContext> stack ) {
        sm.setRequestContextStack(stack);
    }
    
    /**
     * Return a content instance according to the scope policy defined for this
     * component.
     */
    public Object getFcContent() throws InstantiationException {
        return sm.getFcContent();
    }
    
    /**
     * Notify the content controller that the specified content instance is no
     * longer needed. If relevant with the the scope policy, this gives the
     * opportunity to the controller to call the {@link Destroy} annotated
     * method on the content instance.
     * 
     * @param content  the content instance which is released
     * @param isEndMethod
     *      true if the method which releases the content instance is annotated
     *      with EndsConversation
     */
    public void releaseFcContent( Object content, boolean isEndMethod ) {
        sm.releaseFcContent(content,isEndMethod);
    }
    
    
    // -------------------------------------------------------------------------
    // Fields and methods required by the mixin class in the base class
    // -------------------------------------------------------------------------

    /**
     * The {@link Controller#initFcController initFcController} method overriden
     * by this mixin.
     *
     * @param ic information about the component to which this controller object
     *      belongs.
     * @throws InstantiationException if the initialization fails.
     */
    public abstract void _super_initFcController(InitializationContext ic)
    throws InstantiationException;
    
    /**
     * The {@link Component} interface of the component to which this controller
     * object belongs.
     */
    public Component _this_weaveableC;

    /**
     * The {@link NameController} interface of the component to which this
     * controller object belongs.
     */
    public NameController _this_weaveableOptNC;

    /**
     * The {@link ReconfigurableComponentContext} interface of the component to
     * which this controller object belongs.
     */
    public ReconfigurableComponentContext _this_weavableCompCtx;
    
    /**
     * The {@link SCAPropertyController} interface of the component to which
     * this controller object belongs.
     */
    public SCAPropertyController _this_weavablePC;

    /**
     * The {@link LifeCycleController} interface of the component to which
     * this controller object belongs.
     */
    public LifeCycleController _this_weaveableOptLC;
}
