/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.control.intent;

import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Method;
import java.util.List;

import org.objectweb.fractal.api.Interface;
import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.control.component.ReconfigurableComponentContext;

/**
 * Implementation of the {@link IntentJoinPoint} interface for introspecting
 * intent join points. An intent join point corresponds to the interface where a
 * service invocation has been intercepted.

 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 0.3
 */
public class IntentJoinPointImpl implements IntentJoinPoint {

    protected List<IntentHandler> handlers;
    protected int index = 0;

    protected ReconfigurableComponentContext intentComponentContext;
    protected Interface intentItf;
    protected Object intentTarget;
    private Method intentMethod;
    protected Object[] intentMethodArguments;
    
    public IntentJoinPointImpl(
        List<IntentHandler> handlers,
        ReconfigurableComponentContext intentComponentContext,
        Interface intentItf,
        Object intentTarget,
        Method intentMethod,
        Object... intentMethodArguments ) {
        
        this.handlers = handlers;
        this.intentComponentContext = intentComponentContext;
        this.intentItf = intentItf;
        this.intentTarget = intentTarget;
        this.intentMethod = intentMethod;
        this.intentMethodArguments = intentMethodArguments;
    }
    
    /**
     * Return the component context for which the intent has been defined.
     */
    public ReconfigurableComponentContext getComponentContext() {
        return intentComponentContext;
    }

    /**
     * Return the component interface (service or reference) for which the
     * intent has been defined.
     * 
     * @since 0.4.5
     */
    public Interface getInterface() {
        return intentItf;
    }
    
    /**
     * Return the method arguments for which the intent has been defined.
     */
    public Object[] getArguments() {
        return intentMethodArguments;
    }

    /**
     * Return the method for which the intent has been defined.
     */
    public Method getMethod() {
        return intentMethod;
    }
    
    /**
     * Returns the static part of this joinpoint.
     *
     * The static part is an accessible object on which a chain of interceptors
     * are installed.
     */
    public AccessibleObject getStaticPart() {
        return intentMethod;
    }
    
    /**
     * Return the object for which the intent has been defined.
     */
    public Object getThis() {
        return intentTarget;
    }

    /**
     * Proceed to the next intent handler.
     */
    public Object proceed() throws Throwable {
        if( index == handlers.size() ) {
            String msg = "Unexpected case: index==handlers.length should have been checked by the subclass";
            throw new TinfiRuntimeException(msg);
        }
        IntentHandler handler = handlers.get(index);
        index++;
        Object ret = handler.invoke(this);
        return ret;
    }
}
